import unittest
import time
import scullery.mqtt
import gc


class VirtualServer(unittest.TestCase):
    def test_loop(self):
        d = [0]

        def p(*x):
            print("Got MQTT response:"+str([x]))
            d[0] += 1

        scullery.mqtt.testCrashOnce = True
        c = scullery.mqtt.get_connection(
            "localhost", 1883, message_bus_name="testBusName")

        time.sleep(0.2)
        c.subscribe("test", p)

        c.publish("test", "someData")

        time.sleep(1)

        self.assertEqual(d[0], 1)

        # Passive connection, all it does is act as a wrapper around the internal message bus, all real MQTT traffic is through
        # the real connection we made above.
        c2 = scullery.mqtt.get_connection("", message_bus_name='testBusName')

        d2 = [0]

        def p2(*x):
            print("Got MQTT response2:"+str([x]))
            d2[0] += 1

        time.sleep(0.2)
        c2.subscribe("test", p2)

        c2.publish("test", "someData2")

        time.sleep(0.2)

        self.assertEqual(d2[0], 1)

        c2.close()
        c.close()

        del c2
        del c
        gc.collect()
        gc.collect()

        # Recreate the connection, all subscribers carry over
        c = scullery.mqtt.get_connection(
            "localhost", 1883, message_bus_name="testBusName")
        time.sleep(1)
        c.publish("test", "someData2")
        time.sleep(0.2)
        self.assertEqual(d2[0], 2)

        # Delete the functions. Scullery only weakly references them, so they should just be gone.
        del p
        del p2
        gc.collect()
        gc.collect()

        c.publish("test", "someData2")
        time.sleep(0.2)
        self.assertEqual(d2[0], 2)



import time
from scullery import fluidsynth
import unittest

from scullery import workers
workers.start()


class TestAudio(unittest.TestCase):

    def test_fluidsynth_nojack(self):
        fs = fluidsynth.FluidSynth(jack_client_name="Foo")
        time.sleep(3)
        print("Testing fluidsynth with JACK")
        print("You should hear a MIDI note. This test require's fluidsynth, pyFluidsynth, and it's default sf2 file")
        fs.noteOn(0, 70, 50)
        time.sleep(0.5)
        fs.noteOff(0, 70)
        fs.setInstrument(0, "flute")
        print("You should hear a flute note")
        fs.noteOn(0, 70, 70)
        time.sleep(2)
        fs.noteOff(0, 70)




# Copyright Daniel Dunn 2013
# This file is part of Kaithem Automation.

# Kaithem Automation is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.

# Kaithem Automation is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Kaithem Automation.  If not, see <http://www.gnu.org/licenses/>.

# This file manages a work queue that feeds a threadpool
# Tasks will be performed on a best effort basis and errors will be caught and ignored.

import threading
import sys
import traceback
import logging
import atexit
import time
import collections
import random

from typing import Callable, List, Optional, Any

# I'm reserving the system log for a reasonable low-rate log.
# So this is in a separate namespace that shows up elsewhere.
logger = logging.getLogger("workers")

syslogger = logging.getLogger("system")

lastWorkersError = 0
backgroundFunctionErrorHandlers = []


def testLatency():
    start = time.monotonic()
    x = [0]

    def f():
        x[0] = time.monotonic()

    do(f)

    while time.monotonic() - start < 10:
        time.sleep(0.0001)
        if x[0]:
            return x[0] - start
    raise RuntimeError("No response")


spawnLock = threading.RLock()

maxWorkers = 32
minWorkers = 4

shutdownWait = 60
run = True
taskQueue = collections.deque()


def inWaiting():
    return len(taskQueue)


def waitingtasks():
    "Return the number of tasks in the task queue"
    return len(taskQueue)


def handle_errorInFunction(f):
    print("Error in: " + str(f))


def stop():
    global run
    logging.info("Stopping worker threads")
    run = False


def EXIT():
    # Tell all worker threads to stop and wait for them all to finish.
    # If they aren't finished within the time limit, just exit.
    t = time.time()
    stop()
    for i in workers:
        try:
            # All threads total must be finished within the time limit
            workers[i].join(shutdownWait - (time.time() - t))
            # If we try to exit befoe the thread even has time to start or something
        except RuntimeError:
            pass


workers = {}
workersMutable = {}

wakeupHandles = []
wakeupHandlesMutable = []


def testIntegrity():
    with spawnLock:
        assert len(wakeupHandles) == len(workers)


lastStoppedThread = 0


def _makeWorker(e, q, id, fastMode=False):
    # one worker that just pulls tasks from the queue and does them. Errors are caught and
    # We assume the tasks have their own error stuff
    def workerloop():
        global workers
        global lastStoppedThread
        global wakeupHandles

        f = None

        shouldRun = True

        monotonic = time.monotonic

        lastActivity = monotonic()
        pop = taskQueue.pop

        runningState = [True]
        handle = (e, runningState)
        with spawnLock:
            wakeupHandlesMutable.append(handle)
            wakeupHandles = wakeupHandlesMutable[:]

        while (run):
            try:
                runningState[0] = True
                # While either our direct  queue or the overflow queue has things in it we do them.
                while (len(taskQueue)):
                    try:
                        f = pop()
                    except Exception:
                        f = None

                    if f:
                        try:
                            f[0](*f[1])
                            lastActivity = monotonic()
                        except Exception:
                            global lastWorkersError
                            try:
                                if lastWorkersError < monotonic() - 60:
                                    syslogger.exception(
                                        "Error in function. This message is ratelimited, see debug logs for full.\r\nIn "
                                        + f[0].__name__ + " from " +
                                        f[0].__module__ + "\r\n")
                                    lastWorkersError = monotonic()

                                logger.exception(
                                    "Error in function running in thread pool "
                                    + f[0].__name__ + " from " +
                                    f[0].__module__)
                            except Exception:
                                print("Failed to handle error: " +
                                      traceback.format_exc(6))

                            for i in backgroundFunctionErrorHandlers:
                                try:
                                    i(f)
                                except Exception:
                                    print("Failed to handle error: " +
                                          traceback.format_exc(6))
                        finally:
                            # We do not want f staying around, if might hold references that should be GCed away immediatly
                            f = None
                # Ensure the lock is acqured so we can sto the next time
                # We try in blocking mode
                e.acquire(False)
                runningState[0] = False

                # This check happens *after* setting e.on false, so that if we
                # set it false right after they checked and put something in
                # the queue we get it next loop
                if not taskQueue:
                    # Randomize, so they don't all sync up
                    # FastMode polls at 100Hz
                    x = e.acquire(timeout=(random.random() *
                                           2) if not fastMode else 0.01)
                    runningState[0] = True

                if not x and not taskQueue:
                    if not shouldRun:
                        return

                    # Fast prelim check.
                    if len(workers) > minWorkers:
                        # The elements of handle are never copied anywhere,
                        # Once the list is clear, we can be sure there is no further inserts, and the next round will catch almost
                        # all race conditions. Any remaining one in a million ones will be caught in 1 second
                        if lastActivity < (monotonic() - 10):
                            # This should not block.
                            if spawnLock.acquire(timeout=2):
                                try:
                                    if len(workers) > minWorkers:
                                        # Only stop one thread per 5 seconds to prevent
                                        # chattering

                                        # Also don't stop a thread when there aren't at least
                                        # 2 threads that aren't busy.
                                        unbusyCount = 0
                                        for i in wakeupHandles:
                                            if not i[1][0]:
                                                unbusyCount += 1

                                        if unbusyCount > 2 and lastStoppedThread < (
                                                monotonic() - 5):
                                            lastStoppedThread = monotonic()
                                            shouldRun = None
                                            del workersMutable[id]
                                            workers = workersMutable.copy()
                                            wakeupHandlesMutable.remove(handle)
                                            wakeupHandles = wakeupHandlesMutable[:]
                                finally:
                                    spawnLock.release()
            except Exception:
                print("Exception in worker loop: " + traceback.format_exc(6))

    return workerloop


def addWorker():
    global workers
    if spawnLock.acquire(timeout=60):
        try:
            q = []
            e = threading.Lock()
            e.acquire()

            id = time.time()
            # First worker always polls at 100hz
            t = threading.Thread(target=_makeWorker(e, q, id),
                                 name="nostartstoplog.ThreadPoolWorker-" + str(id))
            workersMutable[id] = t
            t.start()
            workers = workersMutable.copy()
        finally:
            spawnLock.release()
    else:
        raise RuntimeError("Could not get the lock!")


_append = taskQueue.append


def do(func: Callable[..., Any], args: Optional[List[Any]] = None):
    """Run a function in the background

    funct(function):
        A function of 0 arguments to be ran in the background in another thread immediatly,
    """

    args = args or []
    if not callable(func):
        raise ValueError("Non callable value")

    _append((func, args))

    for i in wakeupHandles:
        try:
            if i[0].locked():
                i[0].release()
                return
        except RuntimeError:
            pass

    if len(workers) > 4:
        # Wait and retry before attempting to spawn a new thread, if there is probable already enough.
        # that is a slow problematic thing
        time.sleep(0.001)
        time.sleep(0.0001)

        for i in wakeupHandles:
            try:
                if i[0].locked():
                    i[0].release()
                    return
            except RuntimeError:
                pass

    # Sleep 1/25000th of a second for every item in the queue past the max number of threads
    # In an attempt to rate limit
    if len(taskQueue) > maxWorkers:
        time.sleep(max(0, (len(taskQueue) - maxWorkers) / 25000))

    # No unbusy threads? It must go in the overflow queue.
    # Soft rate limit here should work a bit better than the old hard limit at keeping away
    # the deadlocks.
    # Under lock

    # We also need this fast preliminary check to use that lock as rarely as possible.
    if len(workers) < maxWorkers:
        if spawnLock.acquire(timeout=15):
            try:
                if len(workers) < maxWorkers:
                    addWorker()
                    return
            finally:
                spawnLock.release()
        else:
            print("COULD NOT GET SPAWN LOCK TO CREATE ADDITIONAL THREAD. CONTINUING WITH FEWER THREADS. RESTART SUGGESTED")

    # If we can't spawn a new thread
    # Wait a maximum of 15ms before
    # just giving up and leaving
    # it for when somethin
    # wakes up
    for n in range(0, 25):
        if not taskQueue:
            return
        for i in wakeupHandles:
            try:
                if i[0].locked():
                    i[0].release()
                    return
            except RuntimeError:
                pass
        time.sleep(0.0005)


do_try = do


def start(count=8, qsize=64, shutdown_wait=60):
    global __queue, run, worker_wait, workers, maxWorkers
    run = True
    worker_wait = shutdown_wait

    maxWorkers = count

    syslogger.info("Started worker threads")



from time import sleep
import pexpect
import re
import traceback
import shutil
from os.path import join, expanduser
import tempfile
import logging
import subprocess
from threading import Thread
from .utils import create_daemon
from .utils.log import LOG
import os
import shutil

# evt = re.search(r"received event\: *?\'(.*?)\'", line)

# if evt:
#     self.onDTMF(evt.groups(0))


logging.getLogger("urllib3.connectionpool").setLevel("WARN")
logging.getLogger("pydub.converter").setLevel("WARN")

tmpdir = "/dev/shm/"


class BareSIP(Thread):
    def __init__(self, user, pwd, gateway, tts=None, debug=False, block=True, audiodriver="alsa,default", port=5060):
        self.debug = debug
        self.user = user
        self.pwd = pwd
        self.gateway = gateway

        # Learn the second one when it has to reasign a unique name
        # There is an in and an out client name
        self.jackNames = ['baresip']

        if tts:
            self.tts = tts
        else:
            try:
                from responsive_voice import ResponsiveVoice
                self.tts = ResponsiveVoice(gender=ResponsiveVoice.MALE)
            except ImportError:
                logging.exception(
                    "No responsive_voice module, some features will not work")
        self._login = "sip:{u}:{p}@{g}".format(u=self.user, p=self.pwd,
                                               g=self.gateway)
        self._prev_output = ""
        self.running = False

        # Ready immediately, this is a localhost thing
        self.ready = True
        self.mic_muted = False
        self.abort = False
        self.current_call = None
        self._call_status = None
        self.audio = None
        self._ts = None

        self.jack_port = "baresip"

        cnfdir = os.path.join(tmpdir, "ScullerySIP"+str(port))
        self.cnfdir = cnfdir
        try:
            # Remove any existing
            shutil.rmtree(self.cnfdir)
        except:
            pass

        if not os.path.exists(os.path.join(tmpdir, "ScullerySIP"+str(port))):
            os.mkdir(os.path.join(tmpdir, "ScullerySIP"+str(port)))

        self.cnfdir = cnfdir
        # Using the template, create a configuration dir for
        # the baresip instance we are about to make.
        f = os.path.join(os.path.dirname(__file__), "baresip_template")

        drivers = "module\t\talsa.so"
        if 'jack' in audiodriver:
            drivers = "module\t\tjack.so"

        for i in os.listdir(f):
            with open(os.path.join(f, i)) as fd:
                x = fd.read()

            x = x.replace("USERNAME", user)
            x = x.replace("AUDIODRIVER", audiodriver)
            x = x.replace("PORT", str(port))
            x = x.replace("DRIVERS", str(drivers))

            with open(os.path.join(cnfdir, i), "w") as fd:
                fd.write(x)

        self.baresip = pexpect.spawn('baresip', ["-f", cnfdir])
        super().__init__()
        self.daemon = True

        self.start()
        if block:
            self.wait_until_ready()

    # properties
    @property
    def call_established(self):
        return self.call_status == "ESTABLISHED"

    @property
    def call_status(self):
        return self._call_status or "Dis_connected"

    # actions
    def do_command(self, action):
        if self.ready:
            action = str(action)
            self.baresip.sendline(action)
        else:
            LOG.warning(action + " not executed!")
            LOG.exception("NOT READY! please wait")

    def login(self):
        LOG.info("Adding account: " + self.user)
        self.baresip.sendline("/uanew " + self._login)

    def call(self, number):
        LOG.info("Dialling: " + number)
        self.do_command("/dial " + number)

    def hang(self):
        if self.current_call:
            LOG.info("Hanging: " + self.current_call)
            self.do_command("/hangup")
            self.current_call = None
            self._call_status = None
        else:
            LOG.error("No active call to hang")

    def hold(self):
        if self.current_call:
            LOG.info("Holding: " + self.current_call)
            self.do_command("/hold")
        else:
            LOG.error("No active call to hold")

    def resume(self):
        if self.current_call:
            LOG.info("Resuming " + self.current_call)
            self.do_command("/resume")
        else:
            LOG.error("No active call to resume")

    def mute_mic(self):
        if not self.call_established:
            LOG.error("Can not mute microphone while not in a call")
            return
        if not self.mic_muted:
            LOG.info("Muting mic")
            self.do_command("/mute")
        else:
            LOG.info("Mic already muted")

    def unmute_mic(self):
        if not self.call_established:
            LOG.error("Can not unmute microphone while not in a call")
            return
        if self.mic_muted:
            LOG.info("Unmuting mic")
            self.do_command("/mute")
        else:
            LOG.info("Mic already unmuted")

    def accept_call(self):
        self.do_command("/accept")
        status = "ESTABLISHED"
        self.handle_call_status(status)
        self._call_status = status

    def list_calls(self):
        self.do_command("/listcalls")

    def check_call_status(self):
        self.do_command("/callstat")
        sleep(0.1)
        return self.call_status

    def quit(self):
        LOG.info("Exiting")
        if self.running:
            if self.current_call:
                self.hang()
            self.baresip.sendline("/quit")
        self.running = False
        self.current_call = None
        self._call_status = None
        self.abort = True

    def send_dtmf(self, number):
        from opentone import ToneGenerator
        number = str(number)
        for n in number:
            if n not in "0123456789":
                LOG.error("invalid dtmf tone")
                return
        LOG.info("Sending dtmf tones for " + number)
        dtmf = join(tempfile.gettempdir(), "DTMF"+number + ".wav")
        ToneGenerator().dtmf_to_wave(number, dtmf)
        self.send_audio(dtmf)

    def speak(self, speech):
        if not self.call_established:
            LOG.error("Speaking without an active call!")
        else:
            LOG.info("Sending TTS for " + speech)
            self.send_audio(self.tts.get_mp3(speech))
            sleep(0.5)

    def send_audio(self, wav_file):
        if not self.call_established:
            LOG.error("Can't send audio without an active call!")
            return
        wav_file, duration = self.convert_audio(wav_file)
        # send audio stream
        LOG.info("transmitting audio")
        self.do_command("/ausrc aufile," + wav_file)
        # wait till playback ends
        sleep(duration - 0.5)
        # avoid baresip exiting
        self.do_command("/ausrc alsa,default")

    @staticmethod
    def convert_audio(input_file, outfile=None):
        from pydub import AudioSegment
        input_file = expanduser(input_file)
        sound = AudioSegment.from_file(input_file)
        sound += AudioSegment.silent(duration=500)
        # ensure minimum time
        # workaround baresip bug
        while sound.duration_seconds < 3:
            sound += AudioSegment.silent(duration=500)

        outfile = outfile or join(tempfile.gettempdir(), "pybaresip.wav")
        sound = sound.set_frame_rate(48000)
        sound = sound.set_channels(2)
        sound.export(outfile, format="wav")
        return outfile, sound.duration_seconds

    # this is played out loud over speakers
    def say(self, speech):
        if not self.call_established:
            LOG.warning("Speaking without an active call!")
        self.tts.say(speech, blocking=True)

    def play(self, audio_file, blocking=True):
        if not audio_file.endswith(".wav"):
            audio_file, duration = self.convert_audio(audio_file)
        self.audio = self._play_wav(audio_file, blocking=blocking)

    def stop_playing(self):
        if self.audio is not None:
            self.audio.kill()

    @staticmethod
    def _play_wav(wav_file, play_cmd="aplay %1", blocking=False):
        play_mp3_cmd = str(play_cmd).split(" ")
        for index, cmd in enumerate(play_mp3_cmd):
            if cmd == "%1":
                play_mp3_cmd[index] = wav_file
        if blocking:
            return subprocess.call(play_mp3_cmd)
        else:
            return subprocess.Popen(play_mp3_cmd)

    # events
    def handle_incoming_call(self, number):
        LOG.info("Incoming call: " + number)
        if self.call_established:
            LOG.info("already in a call, rejecting")
            sleep(0.1)
            self.do_command("b")
        else:
            LOG.info("default behaviour, rejecting call")
            sleep(0.1)
            self.do_command("b")

    def handle_call_rejected(self, number):
        LOG.info("Rejected incoming call: " + number)

    def handle_call_timestamp(self, timestr):
        LOG.info("Call time: " + timestr)

    def handle_call_status(self, status):
        if status != self._call_status:
            LOG.debug("Call Status: " + status)

    def handle_call_start(self):
        number = self.current_call
        LOG.info("Calling: " + number)

    def handle_call_ringing(self):
        number = self.current_call
        LOG.info(number + " is Ringing")

    def handle_call_established(self):
        LOG.info("Call established")

    def handle_call_ended(self, reason):
        LOG.info("Call ended")
        LOG.debug("Reason: " + reason)

    def _handle_no_accounts(self):
        LOG.debug("No accounts setup")
        self.login()

    def handle_login_success(self):
        LOG.info("Logged in!")

    def handle_login_failure(self):
        LOG.error("Log in failed!")
        self.quit()

    def handle_ready(self):
        LOG.info("Ready for instructions")

    def handle_mic_muted(self):
        LOG.info("Microphone muted")

    def handle_mic_unmuted(self):
        LOG.info("Microphone unmuted")

    def handle_audio_stream_failure(self):
        LOG.debug("Aborting call, maybe we reached voicemail?")
        self.hang()

    def handle_error(self, error):
        LOG.error(error)
        if error == "failed to set audio-source (No such device)":
            self.handle_audio_stream_failure()

    def onJackAssigned(self, cname):
        if not cname in self.jackNames:
            self.jackNames.append(cname)
            # We only have 2, if there are more,
            # it means the default has been reassigned
            if len(self.jackNames) > 2:
                self.jackNames.pop(0)

    def on_audio_rtp(self, rtp):
        pass

    # event loop
    def run(self):
        self.running = True
        while self.running:
            try:
                out = self.baresip.readline().decode("utf-8")

                if out != self._prev_output:
                    out = out.strip()
                    if self.debug:
                        LOG.debug(out)
                    if "baresip is ready." in out:
                        self.handle_ready()
                    elif "account: No SIP accounts found" in out:
                        self._handle_no_accounts()
                    elif "All 1 useragent registered successfully!" in out:
                        self.ready = True
                        self.handle_login_success()

                    elif "ua: SIP register failed:" in out or\
                            "401 Unauthorized" in out or \
                            "Register: Destination address required" in out or\
                            "Register: Connection timed out" in out:
                        self.handle_error(out)
                        self.handle_login_failure()
                    elif "Incoming call from: " in out:
                        num = out.split("Incoming call from: ")[
                            1].split(" - (press 'a' to accept)")[0].strip()
                        self.current_call = num
                        self._call_status = "INCOMING"
                        self.handle_incoming_call(num)
                    elif "call: rejecting incoming call from " in out:
                        num = out.split("rejecting incoming call from ")[
                            1].split(" ")[0].strip()
                        self.handle_call_rejected(num)
                    elif "call: SIP Progress: 180 Ringing" in out:
                        self.handle_call_ringing()
                        status = "RINGING"
                        self.handle_call_status(status)
                        self._call_status = status
                    elif "call: connecting to " in out:
                        n = out.split("call: connecting to '")[1].split("'")[0]
                        self.current_call = n
                        self.handle_call_start()
                        status = "OUTGOING"
                        self.handle_call_status(status)
                        self._call_status = status
                    elif "Call established:" in out:

                        status = "ESTABLISHED"
                        self.handle_call_status(status)
                        self._call_status = status
                        sleep(0.5)
                        self.handle_call_established()
                    elif "call: hold " in out:
                        n = out.split("call: hold ")[1]
                        status = "ON HOLD"
                        self.handle_call_status(status)
                        self._call_status = status
                    elif "Call with " in out and \
                            "terminated (duration: " in out:
                        status = "Dis_connected"
                        duration = out.split("terminated (duration: ")[1][:-1]
                        self.handle_call_status(status)
                        self._call_status = status
                        self.handle_call_timestamp(duration)
                        self.mic_muted = False
                    elif "call muted" in out:
                        self.mic_muted = True
                        self.handle_mic_muted()
                    elif "call un-muted" in out:
                        self.mic_muted = False
                        self.handle_mic_unmuted()
                    elif "session closed:" in out:
                        reason = out.split("session closed:")[1].strip()
                        status = "Dis_connected"
                        self.handle_call_status(status)
                        self._call_status = status
                        self.handle_call_ended(reason)
                        self.mic_muted = False
                    elif "(no active calls)" in out:
                        status = "Dis_connected"
                        self.handle_call_status(status)
                        self._call_status = status
                    elif "incoming rtp for 'audio' established, receiving from " in out:
                        rtp = out.split(
                            "stream: incoming rtp for 'audio' established, receiving from ")
                        self.on_audio_rtp(rtp)

                    elif "===== Call debug " in out:
                        status = out.split("(")[1].split(")")[0]
                        self.handle_call_status(status)
                        self._call_status = status

                    elif 'jack' in out:
                        match = re.search(
                            r"jack: unique name \`(.*?)\' assigned", out)
                        if match:
                            self.onJackAssigned(match.groups(1)[0])

                    elif "--- List of active calls (1): ---" in \
                            self._prev_output:
                        if "ESTABLISHED" in out and self.current_call in out:
                            ts = out.split("ESTABLISHED")[0].split(
                                "[line 1]")[1].strip()
                            if ts != self._ts:
                                self._ts = ts
                                self.handle_call_timestamp(ts)

                    elif "failed to set audio-source (No such device)" in out:
                        error = "failed to set audio-source (No such device)"
                        self.handle_error(error)

                    self._prev_output = out
            except pexpect.exceptions.EOF:
                # baresip exited
                self.quit()
            except pexpect.exceptions.TIMEOUT:
                # nothing happened for a while
                pass
            except Exception:
                print(traceback.format_exc())

    def wait_until_ready(self):
        while not self.ready:
            sleep(0.1)
            if self.abort:
                return




# Copyright Daniel Dunn 2020

# Scullery is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.

# Scullery is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Scullery.  If not, see <http://www.gnu.org/licenses/>.

import time
import weakref
import os
import yaml
import threading
import logging
gmInstruments = None

players = weakref.WeakValueDictionary()

lock = threading.Lock()


def all_notes_off():
    try:
        for i in players:
            players[i].fs.all_notes_off(-1)
    except:
        pass


def stop_all():
    try:
        for i in players:
            players[i].close()
    except:
        pass


def remake_all():
    try:
        for i in players:
            players[i].close()
    except:
        pass


def get_gm_instruments():
    global gmInstruments
    if gmInstruments:
        return gmInstruments
    with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'gm_instruments.yaml')) as f:
        gmInstruments = yaml.load(f.read(), yaml.SafeLoader)
    return gmInstruments


def find_gm_instrument(name, look_in_soundfont=None, bank=None):
    # Allow manually selected instruments
    try:
        return (bank, int(name))
    except:
        pass
    name = name.replace("(", '').replace(')', '')
    # Try to find a matching patch name in the soundfont
    if look_in_soundfont:
        try:
            from sf2utils.sf2parse import Sf2File
            with open(look_in_soundfont, 'rb') as sf2_file:
                sf2 = Sf2File(sf2_file)
            # Names
            x = [i[0].split(b"\0")[0].decode("utf8")
                 for i in sf2.raw.pdta['Phdr']]
        except Exception as e:
            logging.exception("Can't get metadata from this soundfont")
            print("Error looking through soundfont data", e)
            x = []

        # Indexes
        for i in range(len(x)):
            n = x[i].lower()
            n = n.replace("(", '').replace(')', '')
            n = n.split(" ")
            match = True

            for j in name.lower().split(" "):
                if not j in n:
                    # Bank 128 is reserved for drums so it can substitute for drum related words,
                    # It's still a match
                    if not (j in ('kit', 'drums', 'drum') and sf2.raw.pdta['Phdr'][i][2] == 128):
                        match = False
            if match:
                if bank == None:
                    return (sf2.raw.pdta['Phdr'][i][2], sf2.raw.pdta['Phdr'][i][1])
                return (bank, sf2.raw.pdta['Phdr'][i][1])

    x = get_gm_instruments()
    for i in x:
        n = x[i].lower()
        n = n.replace("(", '').replace(')', '')
        n = n.split(" ")
        match = True
        for j in name.lower().split(" "):
            if not j in n:
                match = False
        if match:
            return (bank or 0, i)
    raise ValueError("No matching instrument")


def wait_for_jack():
    from scullery import jacktools
    for i in range(10):
        if not jacktools.get_ports():
            time.sleep(1)
        else:
            return

    raise RuntimeError("It appears that JACK is not running")


def find_sound_font(specific=None, extra_fallback=None):
    # Support the debian, Arch, and EmberOS conventions
    l = ['/usr/share/sounds/sf3/MuseScore_General_Lite.sf3',
         '/var/public.files/emberos/SoundFonts/MuseScore_General.sf3',
         '/usr/share/sounds/sf3/MuseScore_General.sf3',
         "/usr/share/sounds/sf2/FluidR3_GM.sf2",
         '/var/public.files/emberos/SoundFonts/FluidR3_GM.sf3',
         '/usr/share/soundfonts/FluidR3_GM.sf2'
         ]

    if os.path.exists(specific):
        return specific

    for i in l:
        if os.path.exists(l):
            return l

    return extra_fallback


class FluidSynth():
    defaultSoundfont = "/usr/share/sounds/sf2/FluidR3_GM.sf2"

    def __init__(self, soundfont=None, jack_client_name=None, gain=0.2,
                 connect_midi=None, connect_output=None, reverb=False, chorus=False, ondemand=True):
        players[id(self)] = self

        if jack_client_name:
            from . import jacktools
            wait_for_jack()

        self.soundfont = soundfont or self.defaultSoundfont

        if not os.path.isfile(self.soundfont):
            raise OSError("Soundfont: "+soundfont +
                          " does not exist or is not a file")

        from . thirdparty import fluidsynth

        def remake():
            self.fs = fluidsynth.Synth(gain=gain)
            self.fs.setting("synth.chorus.active", 1 if chorus else 0)
            self.fs.setting("synth.reverb.active", 1 if reverb else 0)

            try:
                self.fs.setting("synth.dynamic-sample-loading",
                                1 if ondemand else 0)
            except:
                logging.exception("No dynamic loading support, ignoring")
            self.sfid = self.fs.sfload(self.soundfont)
            usingJack = False

            if jack_client_name:
                self.fs.setting("audio.jack.id", jack_client_name)
                self.fs.setting("midi.jack.id", "fstest")

                usingJack = True

            if connect_midi:
                pass
                # self.midiAirwire = jackmanager.Mono

            if connect_output:
                self.airwire = jacktools.Airwire(
                    jack_client_name or 'KaithemFluidsynth', connect_output)
                self.airwire.connect()

            if usingJack:
                if not jack_client_name:
                    self.fs.setting("audio.jack.id", "fstest")
                    self.fs.setting("midi.jack.id", "fstest")

                self.fs.setting("midi.driver", 'jack')
                self.fs.start(driver="jack", midi_driver="jack")

            else:
                # self.fs.setting("audio.driver", 'alsa')
                self.fs.start()
            for i in range(16):
                self.fs.program_select(i, self.sfid, 0, 0)
        remake()

        # allow restart after JACK settings change
        self.remake = remake

    def setInstrument(self, channel, instrument, bank=None):
        bank, insNumber = find_gm_instrument(instrument, self.soundfont, bank)
        self.fs.program_select(channel, self.sfid, bank, insNumber)

    def noteOn(self, channel, note, velocity):
        self.fs.noteon(channel, note, velocity)

    def noteOff(self, channel, note):
        self.fs.noteoff(channel, note)

    def cc(self, channel, control, val):
        self.fs.cc(channel, control, val)

    def pitchBend(self, channel, val):
        self.fs.pitch_bend(channel, val)

    def programChange(self, channel, val):
        self.fs.program_change(channel, val)

    def __del__(self):
        self.close()

    def close(self):
        with lock:
            try:
                if hasattr(self, 'fs'):
                    self.fs.delete()
                    del self.fs
            except AttributeError:
                pass





from scullery import messagebus
import unittest
import gc
import time
import os
import random

from scullery import workers
workers.start()


def workerTest(extra=0):
    l = []

    def f():
        l.append(1)
    workers.do(f)
    for i in range(200000):
        if l:
            break
        time.sleep(0.00001)
    time.sleep(extra)
    return len(l)


def workerSpam(count=10):
    l = []

    def f():
        l.append(1)
    for i in range(count):
        workers.do(f)


assert (workerTest())


class TestMsgbus(unittest.TestCase):
    def test_sync(self):
        p = []

        def f(t, v):
            p.append((t, v))
        messagebus.subscribe("/test", f)
        messagebus.post_message("/test", "foo", synchronous=True)

        self.assertEqual(p, [('/test', 'foo')])

    def test_del(self):
        "Test unsubscribing by deleting a function"
        p = []

        def f(v):
            p.append(v)
        messagebus.subscribe("test", f)
        del f
        gc.collect()

        messagebus.post_message("test", "foo", synchronous=True)

        self.assertEqual(p, [])

    def test_async(self):
        p = []

        def f(v):
            p.append(v)
        messagebus.subscribe("test", f)
        messagebus.post_message("test", "foo")

        s = 100
        while (not p) and s:
            time.sleep(0.01)
            s -= 1
        self.assertEqual(p, ['foo'])

    def test_Workers(self):
        workerSpam(100)
        self.assertLessEqual(len(workers.workers), workers.maxWorkers)

        for i in range(1000):
            self.assertEqual(workerTest(), 1)

        for i in range(100):
            self.assertEqual(workerTest(), 1)
            time.sleep(0.001)

        for i in range(1):
            self.assertEqual(workerTest(0.1), 1)

        for i in range(100):
            self.assertEqual(workerTest(), 1)
            time.sleep(random.random()*0.001)

        self.assertLessEqual(len(workers.workers), workers.maxWorkers)

        time.sleep(8)
        # Workers should stop within 1s of inactivity.
        self.assertEqual(len(workers.workers), workers.minWorkers)



import gc
import random
import unittest
import time
import os
import scullery.jacktools
import scullery.iceflow
import scullery.workers
scullery.workers.start()


class Player(scullery.iceflow.GstreamerPipeline):
    def __init__(self):
        scullery.iceflow.GstreamerPipeline.__init__(self, realtime=False)
        self.sink = self.add_element('audiotestsrc')
        self.sink = self.add_element('jackaudiosink', client_name="JackTest")


# scullery.jacktools.start_managing()

# We call this manually, because we didn't set manageJackProcess
# So it won't actually auto manage the server itself
# scullery.jacktools._startJackProcess()

class TestJackAudio(unittest.TestCase):
    def test_airwire(self):
        p = Player()
        p.start()

        print("You should hear noise")
        aw = scullery.jacktools.Airwire("JackTest", "system")
        aw.connect()
        time.sleep(1)
        print("No more noise")

        del aw
        gc.collect()

        self.assertEqual(len(scullery.jacktools.allConnections), 0)

    def test_fluidsynth_jack(self):
        import scullery.fluidsynth
        fs = scullery.fluidsynth.FluidSynth(
            jack_client_name="fstest", connect_output="system")

        print("You should hear a MIDI note. This test require's fluidsynth, pyFluidsynth, and it's default sf2 file")
        fs.noteOn(0, 70, 50)
        time.sleep(0.5)
        fs.noteOff(0, 70)
        fs.setInstrument(0, "Pizzicato")
        print("You should hear a Pizzicato string note")
        fs.noteOn(0, 70, 70)
        time.sleep(1)
        fs.noteOff(0, 70)

    def test_z_stop(self):
        "Must obviously run last"
        print("Stopping JACK")
        scullery.jacktools.stop_managing()
        scullery.jacktools.stopJackServer()
        print("Stopped JACK")
